package rdwr

import (
	"bytes"
	"context"
	"io"
	"sync"

	"github.com/goccy/go-json"

	"gfx.cafe/open/jrpc/pkg/codec"
	"gfx.cafe/open/jrpc/pkg/serverutil"
	"gfx.cafe/util/go/bytepool"
)

type Codec struct {
	ctx context.Context
	cn  func()

	dec    *json.Decoder
	wrLock sync.Mutex
	wr     *bytes.Buffer
	w      io.Writer
	msgs   chan *serverutil.Bundle
}

func NewCodec(rd io.Reader, wr io.Writer, onError func(error)) *Codec {
	ctx, cn := context.WithCancel(context.TODO())
	c := &Codec{
		ctx:  ctx,
		cn:   cn,
		dec:  json.NewDecoder(rd),
		wr:   new(bytes.Buffer),
		w:    wr,
		msgs: make(chan *serverutil.Bundle, 8),
	}
	return c
}

// gets the peer info
func (c *Codec) PeerInfo() codec.PeerInfo {
	return codec.PeerInfo{
		Transport:  "ipc",
		RemoteAddr: "",
		HTTP:       codec.HttpInfo{},
	}
}

func (c *Codec) ReadBatch(ctx context.Context, fn func([]*codec.Message, bool)) (func(), error) {
	ctx, cn := context.WithCancel(ctx)
	context.AfterFunc(c.ctx, cn)
	var msg json.RawMessage
	msg = bytepool.GetStd()
	err := c.dec.DecodeContext(ctx, &msg)
	if err != nil {
		return nil, err
	}
	return func() {
		defer bytepool.PutStd(msg)
		ans := serverutil.ParseBundle(msg)
		fn(ans.Messages, ans.Batch)
	}, nil
}

// closes the connection
func (c *Codec) Close() error {
	c.cn()
	return nil
}

func (c *Codec) Write(p []byte) (n int, err error) {
	c.wrLock.Lock()
	defer c.wrLock.Unlock()
	return c.wr.Write(p)
}

func (c *Codec) Flush() (err error) {
	c.wrLock.Lock()
	defer c.wrLock.Unlock()
	defer c.wr.Reset()
	err = c.wr.WriteByte('\n')
	if err != nil {
		return err
	}
	_, err = c.wr.WriteTo(c.w)
	if err != nil {
		return err
	}
	return nil
}

// Closed returns a channel which is closed when the connection is closed.
func (c *Codec) Closed() <-chan struct{} {
	return c.ctx.Done()
}

// RemoteAddr returns the peer address of the connection.
func (c *Codec) RemoteAddr() string {
	return ""
}

// Dialrdwr attaches an in-process connection to the given RPC server.
// func Dialrdwr(handler *Server) *Client {
//	initctx := context.Background()
//	c, _ := newClient(initctx, func(context.Context) (ServerCodec, error) {
//		p1, p2 := net.Pipe()
//		go handler.ServeCodec(NewCodec(p1))
//		return NewCodec(p2), nil
//	})
//	return c
// }
